# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""End2end tests base class."""

import json
import logging
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import pytest
from aea.configurations.base import PublicId
from aea.test_tools.test_cases import AEATestCaseMany, Result
from aea_test_autonomy.configurations import ANY_ADDRESS
from aea_test_autonomy.docker.registries import SERVICE_REGISTRY
from aea_test_autonomy.fixture_helpers import (  # noqa: F401; pylint: disable=unused-import
    FlaskTendermintDockerImage,
    UseFlaskTendermintNode,
    UseLocalIpfs,
)
from web3 import Web3


TERMINATION_TIMEOUT = 120
_HTTP = "http://"


@dataclass
class RoundChecks:
    """
    Class for the necessary checks of a round during the tests.

    name: is the name of the round for which the checks should be performed.
    event: is the name of the event that is considered as successful.
    n_periods: is the number of periods this event should appear for the check to be considered successful.
    """

    name: str
    success_event: str = "DONE"
    n_periods: int = 1


@pytest.mark.e2e
@pytest.mark.integration
class BaseTestEnd2End(AEATestCaseMany, UseFlaskTendermintNode, UseLocalIpfs):
    """
    Base class for end-to-end tests of agents with a skill extending the abstract_abci_round skill.

    The setup test function of this class will configure a set of 'n'
    agents with the configured (agent_package) agent, and a Tendermint network
    of 'n' nodes, one for each agent.

    Test subclasses must set `agent_package`, `wait_to_finish` and `check_strings`.
    """

    # where to fetch agent from
    IS_LOCAL = True
    # whether to capture logs of subprocesses
    capture_log = True
    # generic service configurations
    ROUND_TIMEOUT_SECONDS = 10.0
    RESET_PAUSE_DURATION = 10
    KEEPER_TIMEOUT = 30.0
    # generic node healthcheck configuration
    HEALTH_CHECK_MAX_RETRIES = 20
    HEALTH_CHECK_SLEEP_INTERVAL = 3.0
    # log option for agent process
    cli_log_options = ["-v", "DEBUG"]
    # reserved variable for subprocesses collection
    processes: Dict
    # name of agent and skill (the one with composed abci apps) to be tested
    agent_package: str
    skill_package: str
    # time to wait for testable log strings to appear
    wait_to_finish: int
    # the "happy path" is a successful finish of the FSM execution
    happy_path: Tuple[RoundChecks, ...] = ()
    # tuple of strings expected to appear in output as is.
    strict_check_strings: Tuple[str, ...] = ()
    # process to be excluded from checks
    exclude_from_checks: List[int] = []
    # dictionary of extra config overrides to be applied
    extra_configs: List[Dict[str, Any]] = []
    # ledger used for testing
    ledger_id: str = "ethereum"
    key_file_name: str = "ethereum_private_key.txt"
    USE_GRPC = False
    # usually test envs provide the full set of dependencies; only select
    # when you want to run the installation of the packages as part of the test case.
    RUN_AEA_INSTALL = False

    @classmethod
    def setup_class(cls) -> None:
        """Setup class"""

        FlaskTendermintDockerImage.use_grpc = cls.USE_GRPC  # wicked
        super().setup_class()

    @classmethod
    def set_config(
        cls,
        dotted_path: str,
        value: Any,
        type_: Optional[str] = None,
        aev: bool = True,
    ) -> Result:
        """Set config value."""
        return super().set_config(dotted_path, value, type_, aev)

    def __set_extra_configs(self) -> None:
        """Set the current agent's extra config overrides that are skill specific."""
        for config in self.extra_configs:
            self.set_config(**config)

    def __set_configs(self, i: int, nb_agents: int) -> None:
        """Set the current agent's config overrides."""
        # each agent has its Tendermint node instance

        skill = PublicId.from_str(self.skill_package)

        self.set_config(
            "agent.logging_config.handlers.logfile.filename",
            str(self.t / f"abci_{i}.txt"),
        )
        self.set_config(
            "vendor.valory.connections.abci.config.host",
            ANY_ADDRESS,
        )
        self.set_config(
            "vendor.valory.connections.abci.config.port",
            self.get_abci_port(i),
        )
        self.set_config(
            "vendor.valory.connections.abci.config.use_tendermint",
            False,
        )
        self.set_config("vendor.valory.connections.abci.config.use_grpc", self.USE_GRPC)
        self.set_config(
            "vendor.valory.connections.abci.config.tendermint_config.rpc_laddr",
            self.get_laddr(i),
        )
        self.set_config(
            "vendor.valory.connections.abci.config.tendermint_config.p2p_laddr",
            self.get_laddr(i, p2p=True),
        )
        self.set_config(
            "vendor.valory.connections.abci.config.tendermint_config.p2p_seeds",
            json.dumps(self.p2p_seeds),
            "list",
        )
        key_pairs = getattr(self, "key_pairs", None)
        if key_pairs is not None:
            self.set_config(
                f"vendor.{skill.author}.skills.{skill.name}.models.params.args.setup.all_participants",
                json.dumps(
                    [
                        Web3.toChecksumAddress(pairs[0])
                        for pairs in key_pairs[:nb_agents]
                    ]
                ),
                "list",
            )
        self.set_config(
            f"vendor.{skill.author}.skills.{skill.name}.models.params.args.reset_tendermint_after",
            5,
        )
        self.set_config(
            f"vendor.{skill.author}.skills.{skill.name}.models.params.args.round_timeout_seconds",
            self.ROUND_TIMEOUT_SECONDS,
            type_="float",
        )
        self.set_config(
            f"vendor.{skill.author}.skills.{skill.name}.models.params.args.tendermint_url",
            f"{_HTTP}{ANY_ADDRESS}:{self.get_port(i)}",
        )
        self.set_config(
            f"vendor.{skill.author}.skills.{skill.name}.models.params.args.tendermint_com_url",
            f"{_HTTP}{ANY_ADDRESS}:{self.get_com_port(i)}",
        )
        self.set_config(
            f"vendor.{skill.author}.skills.{skill.name}.models.params.args.keeper_timeout",
            self.KEEPER_TIMEOUT,
            type_="float",
        )
        self.set_config(
            f"vendor.{skill.author}.skills.{skill.name}.models.benchmark_tool.args.log_dir",
            str(self.t),
            type_="str",
        )
        self.set_config(
            f"vendor.{skill.author}.skills.{skill.name}.models.params.args.reset_pause_duration",
            self.RESET_PAUSE_DURATION,
            type_="int",
        )
        self.set_config(
            f"vendor.{skill.author}.skills.{skill.name}.models.params.args.service_registry_address",
            SERVICE_REGISTRY,  # address on registries image
            type_="str",
        )
        self.set_config(  # dummy service
            f"vendor.{skill.author}.skills.{skill.name}.models.params.args.on_chain_service_id",
            "1",
            type_="int",
        )
        self.set_config(
            "vendor.valory.connections.ipfs.config.ipfs_domain",
            self.ipfs_domain,
        )
        self.__set_extra_configs()

    @staticmethod
    def _get_agent_name(i: int) -> str:
        """Get the ith agent's name."""
        return f"agent_{i:05d}"

    def __prepare_agent_i(self, i: int, nb_agents: int) -> None:
        """Prepare the i-th agent."""
        agent_name = self._get_agent_name(i)
        logging.info(f"Processing agent {agent_name}...")
        self.fetch_agent(self.agent_package, agent_name, is_local=self.IS_LOCAL)
        self.set_agent_context(agent_name)
        if hasattr(self, "key_pairs"):
            Path(  # pylint: disable=unspecified-encoding
                self.current_agent_context, self.key_file_name
            ).write_text(
                self.key_pairs[i][1]  # type: ignore
            )
        else:
            self.generate_private_key(self.ledger_id, self.key_file_name)
        self.add_private_key(self.ledger_id, self.key_file_name)
        self.__set_configs(i, nb_agents)
        # issue certificates for libp2p proof of representation
        self.generate_private_key("cosmos", "cosmos_private_key.txt")
        self.add_private_key("cosmos", "cosmos_private_key.txt")
        self.run_cli_command("issue-certificates", cwd=self._get_cwd())

    def prepare(self, nb_nodes: int) -> None:
        """Set up the agents."""
        for agent_id in range(nb_nodes):
            self.__prepare_agent_i(agent_id, nb_nodes)

        if self.RUN_AEA_INSTALL:
            # run 'aea install' in only one AEA project, to save time
            self.set_agent_context(self._get_agent_name(0))
            self.run_install()

    def prepare_and_launch(self, nb_nodes: int) -> None:
        """Prepare and launch the agents."""
        self.processes = dict.fromkeys(range(nb_nodes))
        self.prepare(nb_nodes)
        for agent_id in range(nb_nodes):
            self._launch_agent_i(agent_id)

    def _launch_agent_i(self, i: int) -> None:
        """Launch the i-th agent."""
        agent_name = self._get_agent_name(i)
        logging.info(f"Launching agent {agent_name}...")
        self.set_agent_context(agent_name)
        process = self.run_agent()
        self.processes[i] = process

    @staticmethod
    def __generate_full_strings_from_rounds(
        happy_path: Tuple[RoundChecks, ...]
    ) -> Dict[str, int]:
        """Generate the full strings from the given round strings"""
        full_strings = {}
        for round_check in happy_path:
            entered_str = f"Entered in the '{round_check.name}' round for period"
            done_str = f"'{round_check.name}' round is done with event: Event.{round_check.success_event}"
            full_strings[entered_str] = round_check.n_periods
            full_strings[done_str] = round_check.n_periods

        return full_strings

    @classmethod
    def missing_from_output(  # type: ignore  # pylint: disable=arguments-differ
        cls,
        happy_path: Tuple[RoundChecks, ...] = (),
        strict_check_strings: Tuple[str, ...] = (),
        period: int = 1,
        is_terminating: bool = True,
        **kwargs: Any,
    ) -> Tuple[List[str], List[str]]:
        """
        Check if strings are present in process output.

        Read process stdout in thread and terminate when all strings are present or timeout expired.

        :param happy_path: the happy path of the testing FSM.
        :param strict_check_strings: tuple of strings expected to appear in output as is.
        :param period: period of checking.
        :param is_terminating: whether the agents are terminated if any of the check strings do not appear in the logs.
        :param kwargs: the kwargs of the overridden method.
        :return: tuple with two lists of missed strings, the strict and the round respectively.
        """
        # Call the original method with the strict checks.
        kwargs["strings"] = strict_check_strings
        kwargs["is_terminating"] = False
        missing_strict_strings = super().missing_from_output(**kwargs)

        # Perform checks for the round strings.
        missing_round_strings = []
        if len(happy_path) > 0:
            logging.info("Performing checks for the round strings.")
            check_strings_to_n_periods = cls.__generate_full_strings_from_rounds(
                happy_path
            )
            # Create dictionary to keep track of how many times this string has appeared so far.
            check_strings_to_n_appearances = dict.fromkeys(
                check_strings_to_n_periods, 0
            )
            end_time = time.time() + kwargs["timeout"]
            # iterate while the check strings are still present in the dictionary,
            # i.e. have not appeared for the required amount of times.
            while bool(check_strings_to_n_periods) and time.time() < end_time:
                for line in check_strings_to_n_periods.copy().keys():
                    # count the number of times the line has appeared so far.
                    n_times_appeared: int = cls.stdout[kwargs["process"].pid].count(
                        line
                    )
                    # track the number times the line has appeared so far.
                    check_strings_to_n_appearances[line] = n_times_appeared
                    # if the required number has been reached, delete them from the check dictionaries.
                    if (
                        check_strings_to_n_appearances[line]
                        >= check_strings_to_n_periods[line]
                    ):
                        del check_strings_to_n_periods[line]
                        del check_strings_to_n_appearances[line]

                # sleep for `period` amount of time.
                time.sleep(period)

            # generate the missing strings with the number of times they are missing.
            missing_round_strings = [
                f"'{s}' appeared only {n_appeared} out of {n_expected} times"
                for (s, n_expected), (_, n_appeared) in zip(
                    check_strings_to_n_periods.items(),
                    check_strings_to_n_appearances.items(),
                )
            ]

        return missing_strict_strings, missing_round_strings

    def __check_missing_strings(
        self,
        missing_strict_strings: List[str],
        missing_round_strings: List[str],
        i: int,
    ) -> None:
        """Checks for missing strings in agent's output."""
        agent_name = self._get_agent_name(i)
        missing_agent_logs = ""
        if missing_strict_strings:
            missing_agent_logs += f"Strings {missing_strict_strings} didn't appear in {agent_name} output.\n"
        missing_agent_logs += "\n".join(missing_round_strings)

        assert missing_agent_logs == "", missing_agent_logs  # nosec

    def check_aea_messages(self) -> None:
        """
        Check that *each* AEA prints these messages.

        First failing check will cause assertion error and test tear down.
        """
        for i, process in self.processes.items():
            if i in self.exclude_from_checks:
                continue
            (
                missing_strict_strings,
                missing_round_strings,
            ) = self.missing_from_output(
                process=process,
                happy_path=self.happy_path,
                strict_check_strings=self.strict_check_strings,
                timeout=self.wait_to_finish,
            )

            self.__check_missing_strings(
                missing_strict_strings, missing_round_strings, i
            )


class BaseTestEnd2EndExecution(BaseTestEnd2End):
    """
    Test that an agent that is launched later can synchronize with the rest of the network

    - each agent starts, and sets up the ABCI connection, which in turn spawns both an ABCI
      server and a local Tendermint node (using the configuration folders we set up previously).
      The Tendermint node is unique for each agent
    - when we will stop one agent, also the ABCI server created by the ABCI connection will
      stop, and in turn the Tendermint node will stop. In particular, it does not keep polling
      the endpoint until it is up again, it just stops.
    - when we will restart the previously stopped agent, the ABCI connection will set up again
      both the server and the Tendermint node. The node will automatically connect to the rest
      of the Tendermint network, loads the entire blockchain bulit so far by the others, and
      starts sending ABCI requests to the agent (begin_block; deliver_tx*; end_block), plus
      other auxiliary requests like info , flush etc. The agent which is already processing
      incoming messages, forwards the ABCI requests to the ABCIHandler, which produces ABCI
      responses that are forwarded again via the ABCI connection such that the Tendermint
      node can receive the responses
    """

    nb_nodes: int = 0  # number of agents with tendermint nodes

    # configuration specific for agent restart
    stop_string: str  # mandatory argument if n_terminal > 0
    n_terminal: int = 0  # number of agents to be restarted
    wait_to_kill: int = 0  # delay the termination event
    restart_after: int = 60  # how long to wait before restart
    wait_before_stop: int = 15  # how long to check logs for `stop_string`

    def check(self) -> None:
        """Check pre-conditions of the test"""
        if self.n_terminal > self.nb_nodes:
            fail_msg = "Cannot terminate {nb_nodes} out of {n_terminal} agents:"
            pytest.fail(
                fail_msg.format(nb_nodes=self.nb_nodes, n_terminal=self.n_terminal)
            )
        if self.n_terminal and not hasattr(self, "stop_string"):
            pytest.fail("'stop_string' must be provided for agent termination.")

    def test_run(self, nb_nodes: int) -> None:
        """Run the test."""
        self.nb_nodes = nb_nodes
        self.check()

        self.prepare_and_launch(nb_nodes)
        self.health_check(
            max_retries=self.HEALTH_CHECK_MAX_RETRIES,
            sleep_interval=self.HEALTH_CHECK_SLEEP_INTERVAL,
        )
        if self.n_terminal:
            self._restart_agents()

        self.check_aea_messages()
        try:
            self.terminate_agents(timeout=TERMINATION_TIMEOUT)
        except TimeoutError:
            # This is rare, though it has occurred that even after a kill signal
            # with a 30s timeout the subprocess did not terminate.
            logging.error(f"Failed to terminate agents after {TERMINATION_TIMEOUT}s")

    def _restart_agents(self) -> None:
        """Stops and restarts agents after stop string is found."""

        # stop the last agent as soon as the "stop string" is found in the output
        # once found, we start terminating the first agent
        logging.info(f"Waiting for string {self.stop_string} in last agent output")
        missing_strict_strings, _ = self.missing_from_output(
            process=self.processes[max(self.processes)],
            strict_check_strings=(self.stop_string,),
            timeout=self.wait_before_stop,
        )
        if missing_strict_strings:
            msg = f"cannot stop agent, stop string `{self.stop_string}` not found"
            raise RuntimeError(msg)

        if self.wait_to_kill:
            logging.info("Waiting to terminate agents")
            time.sleep(self.wait_to_kill)

        # terminate the agents - sequentially
        # don't pop before termination, seems to lead to failure!
        for i in range(self.n_terminal):
            agent_name = self._get_agent_name(i)
            self.processes[i].terminate()
            self.processes.pop(i)
            logging.info(f"Terminated {agent_name}")

        # wait for some time before restarting
        logging.info(
            f"Waiting {self.restart_after} seconds before restarting the agent"
        )
        time.sleep(self.restart_after)

        # restart agents
        for i in range(self.n_terminal):
            agent_name = self._get_agent_name(i)
            self._launch_agent_i(i)
            logging.info(f"Restarted {agent_name}")

    @classmethod
    def teardown_class(cls) -> None:
        """Teardown the test."""
        super().teardown_class()
        FlaskTendermintDockerImage.cleanup(cls.nb_nodes)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains helper classes/functions for fixtures."""

# needed because many test base classes do not have public methods
# pylint: disable=too-few-public-methods

# needed because many fixture auto-loaders of test classes do not use fixtures
# pylint: disable=unused-argument

import logging
from typing import Any, Dict, Generator, Iterator, List, Tuple, cast

import docker
import pytest
from aea_cli_ipfs.ipfs_utils import IPFSDaemon
from aea_test_autonomy.configurations import GANACHE_CONFIGURATION, KEY_PAIRS, LOCALHOST
from aea_test_autonomy.docker.acn_node import ACNNodeDockerImage, DEFAULT_ACN_CONFIG
from aea_test_autonomy.docker.amm_net import AMMNetDockerImage
from aea_test_autonomy.docker.base import (
    DockerBaseTest,
    DockerImage,
    launch_image,
    launch_many_containers,
)
from aea_test_autonomy.docker.ganache import (
    DEFAULT_GANACHE_ADDR,
    DEFAULT_GANACHE_PORT,
    GanacheDockerImage,
)
from aea_test_autonomy.docker.gnosis_safe_net import (
    DEFAULT_HARDHAT_ADDR,
    DEFAULT_HARDHAT_PORT,
    GnosisSafeNetDockerImage,
)
from aea_test_autonomy.docker.registries import RegistriesDockerImage
from aea_test_autonomy.docker.tendermint import (
    DEFAULT_ABCI_HOST,
    DEFAULT_ABCI_PORT,
    DEFAULT_TENDERMINT_PORT,
    FlaskTendermintDockerImage,
    TendermintDockerImage,
)
from eth_account import Account


logger = logging.getLogger(__name__)


###
# Vanilla Tendermint
###


@pytest.fixture(scope="session")
def tendermint_port() -> int:
    """Get the Tendermint port"""
    return DEFAULT_TENDERMINT_PORT


@pytest.fixture(scope="session")
def abci_host() -> str:
    """Get the ABCI host"""
    return DEFAULT_ABCI_HOST


@pytest.fixture(scope="session")
def abci_port() -> int:
    """Get the ABCI port"""
    return DEFAULT_ABCI_PORT


@pytest.fixture(scope="class")
def tendermint(
    tendermint_port: int,  # pylint: disable=redefined-outer-name
    abci_host: str,  # pylint: disable=redefined-outer-name
    abci_port: int,  # pylint: disable=redefined-outer-name
    timeout: float = 2.0,
    max_attempts: int = 10,
) -> Generator:
    """Launch the Ganache image."""
    client = docker.from_env()
    logging.info(f"Launching Tendermint at port {tendermint_port}")
    image = TendermintDockerImage(client, abci_host, abci_port, tendermint_port)
    yield from launch_image(image, timeout=timeout, max_attempts=max_attempts)


@pytest.mark.integration
class UseTendermint:
    """Inherit from this class to use Tendermint."""

    _tendermint_image: TendermintDockerImage
    tendermint_port: int

    @pytest.fixture(autouse=True, scope="class")
    def _start_tendermint(
        self,
        tendermint: TendermintDockerImage,  # pylint: disable=redefined-outer-name
        tendermint_port: int,  # pylint: disable=redefined-outer-name
    ) -> None:
        """Start a Tendermint image."""
        cls = type(self)
        cls._tendermint_image = tendermint  # pylint: disable=protected-access
        cls.tendermint_port = tendermint_port

    @property
    def abci_host(self) -> str:
        """Get the abci host address."""
        return self._tendermint_image.abci_host

    @property
    def abci_port(self) -> int:
        """Get the abci port."""
        return self._tendermint_image.abci_port

    @property
    def node_address(self) -> str:
        """Get the node address."""
        return f"http://{LOCALHOST}:{self.tendermint_port}"


###
# Tendermint managed by Flask server
###


@pytest.fixture
def nb_nodes(request: Any) -> int:
    """Get a parametrized number of nodes."""
    return request.param


@pytest.fixture
def flask_tendermint(
    tendermint_port: int,  # pylint: disable=redefined-outer-name
    nb_nodes: int,  # pylint: disable=redefined-outer-name
    abci_host: str,  # pylint: disable=redefined-outer-name
    abci_port: int,  # pylint: disable=redefined-outer-name
    timeout: float = 2.0,
    max_attempts: int = 10,
) -> Generator[FlaskTendermintDockerImage, None, None]:
    """Launch the Flask server with Tendermint container."""
    client = docker.from_env()
    logging.info(
        f"Launching Tendermint nodes managed by Flask server at ports {[tendermint_port + i * 10 for i in range(nb_nodes)]}"
    )
    image = FlaskTendermintDockerImage(client, abci_host, abci_port, tendermint_port)
    yield from cast(
        Generator[FlaskTendermintDockerImage, None, None],
        launch_many_containers(image, nb_nodes, timeout, max_attempts),
    )


@pytest.mark.integration
class UseFlaskTendermintNode:
    """Inherit from this class to use flask server with Tendermint."""

    @pytest.fixture(autouse=True)
    def _start_tendermint(
        self,
        flask_tendermint: FlaskTendermintDockerImage,  # pylint: disable=redefined-outer-name
        tendermint_port: int,  # pylint: disable=redefined-outer-name
    ) -> None:
        """Start a Tendermint image."""
        self._tendermint_image = (  # pylint: disable=attribute-defined-outside-init
            flask_tendermint
        )
        self.tendermint_port = (  # pylint: disable=attribute-defined-outside-init
            tendermint_port
        )

    @property
    def p2p_seeds(self) -> List[str]:
        """Get the p2p seeds."""
        return self._tendermint_image.p2p_seeds

    def get_node_name(self, i: int) -> str:
        """Get the node's name."""
        return self._tendermint_image.get_node_name(i)

    def get_abci_port(self, i: int) -> int:
        """Get the ith rpc port."""
        return self._tendermint_image.get_abci_port(i)

    def get_port(self, i: int) -> int:
        """Get the ith port."""
        return self._tendermint_image.get_port(i)

    def get_com_port(self, i: int) -> int:
        """Get the ith com port."""
        return self._tendermint_image.get_com_port(i)

    def get_laddr(self, i: int, p2p: bool = False) -> str:
        """Get the ith rpc port."""
        return self._tendermint_image.get_addr("tcp://", i, p2p)

    def health_check(self, **kwargs: Any) -> None:
        """Perform a health check."""
        self._tendermint_image.health_check(**kwargs)


###
# Vanilla Ganache
###


@pytest.fixture(scope="session")
def ganache_addr() -> str:
    """HTTP address to the Ganache node."""
    return DEFAULT_GANACHE_ADDR


@pytest.fixture(scope="session")
def ganache_port() -> int:
    """Port of the connection to the Ganache Node to use during the tests."""
    return DEFAULT_GANACHE_PORT


@pytest.fixture(scope="session")
def ganache_configuration() -> Dict:
    """Get the Ganache configuration for testing purposes."""
    return GANACHE_CONFIGURATION


# TODO: remove as not used
@pytest.fixture(scope="function")
def ganache_scope_function(
    ganache_configuration: Dict,  # pylint: disable=redefined-outer-name
    ganache_addr: str,  # pylint: disable=redefined-outer-name
    ganache_port: int,  # pylint: disable=redefined-outer-name
    timeout: float = 2.0,
    max_attempts: int = 10,
) -> Generator:
    """Launch the Ganache image. This fixture is scoped to a function which means it will destroyed at the end of the test."""
    client = docker.from_env()
    image = GanacheDockerImage(
        client, ganache_addr, ganache_port, config=ganache_configuration
    )
    yield from launch_image(image, timeout=timeout, max_attempts=max_attempts)


@pytest.fixture(scope="class")
def ganache_scope_class(
    ganache_configuration: Dict,  # pylint: disable=redefined-outer-name
    ganache_addr: str,  # pylint: disable=redefined-outer-name
    ganache_port: int,  # pylint: disable=redefined-outer-name
    timeout: float = 2.0,
    max_attempts: int = 10,
) -> Generator:
    """Launch the Ganache image. This fixture is scoped to a class which means it will destroyed after running every test in a class."""
    client = docker.from_env()
    image = GanacheDockerImage(
        client, ganache_addr, ganache_port, config=ganache_configuration
    )
    yield from launch_image(image, timeout=timeout, max_attempts=max_attempts)


@pytest.fixture(scope="class")
def ammnet_scope_class(
    timeout: float = 2.0,
    max_attempts: int = 26,
) -> Generator:
    """Launch the Ganache image. This fixture is scoped to a class which means it will destroyed after running every test in a class."""
    client = docker.from_env()
    image = AMMNetDockerImage(client=client)
    yield from launch_image(image, timeout=timeout, max_attempts=max_attempts)


@pytest.mark.integration
class UseGanache:
    """Inherit from this class to use Ganache."""

    key_pairs: List[Tuple[str, str]] = []

    @classmethod
    @pytest.fixture(autouse=True, scope="class")
    def _start_ganache(
        cls,
        ganache_scope_class: GanacheDockerImage,  # pylint: disable=redefined-outer-name
        ganache_configuration: Dict,  # pylint: disable=redefined-outer-name
    ) -> None:
        """Start Ganache instance."""
        cls.key_pairs = cast(
            List[Tuple[str, str]],
            [
                key
                if type(key) == tuple  # pylint: disable=unidiomatic-typecheck
                else (
                    Account.from_key(  # pylint: disable=no-value-for-parameter
                        key
                    ).address,
                    key,
                )
                for key, _ in ganache_configuration.get("accounts_balances", [])
            ],
        )


class GanacheBaseTest(DockerBaseTest):
    """Base pytest class for Ganache."""

    addr: str = DEFAULT_GANACHE_ADDR
    port: int = DEFAULT_GANACHE_PORT
    configuration: Dict = GANACHE_CONFIGURATION

    @classmethod
    def setup_class_kwargs(cls) -> Dict[str, Any]:
        """Get kwargs for _setup_class call."""
        setup_class_kwargs = {
            "key_pairs": cls.key_pairs(),
            "url": cls.url(),
        }
        return setup_class_kwargs

    @classmethod
    def _build_image(cls) -> DockerImage:
        """Build the image."""
        client = docker.from_env()
        return GanacheDockerImage(client, cls.addr, cls.port, config=cls.configuration)

    @classmethod
    def key_pairs(cls) -> List[Tuple[str, str]]:
        """Get the key pairs which are funded."""
        key_pairs_ = cast(
            List[Tuple[str, str]],
            [
                key
                if type(key) == tuple  # pylint: disable=unidiomatic-typecheck
                else (
                    Account.from_key(  # pylint: disable=no-value-for-parameter
                        key
                    ).address,
                    key,
                )
                for key, _ in cls.configuration.get("accounts_balances", [])
            ],
        )
        return key_pairs_

    @classmethod
    def url(cls) -> str:
        """Get the url under which the image is reachable."""
        return f"{cls.addr}:{cls.port}"


###
# Vanilla ACN
###


@pytest.fixture(scope="session")
def acn_config() -> Dict:
    """ACN node configuration."""
    return DEFAULT_ACN_CONFIG


@pytest.fixture(scope="function")
def acn_node(
    acn_config: Dict,  # pylint: disable=redefined-outer-name
    timeout: float = 2.0,
    max_attempts: int = 10,
) -> Generator:
    """Launch the Ganache image."""
    client = docker.from_env()
    logging.info(f"Launching ACNNode with the following config: {acn_config}")
    image = ACNNodeDockerImage(client, acn_config)
    yield from launch_image(image, timeout=timeout, max_attempts=max_attempts)


@pytest.mark.integration
class UseACNNode:
    """Inherit from this class to use an ACNNode for a client connection"""

    _acn_node_image: ACNNodeDockerImage

    @classmethod
    @pytest.fixture(autouse=True)
    def _start_acn(
        cls,
        acn_node: ACNNodeDockerImage,  # pylint: disable=redefined-outer-name
    ) -> None:
        """Start an ACN instance."""
        cls._acn_node_image = acn_node


class ACNNodeBaseTest(DockerBaseTest):
    """Base pytest class for Ganache."""

    configuration: Dict = DEFAULT_ACN_CONFIG

    @classmethod
    def setup_class_kwargs(cls) -> Dict[str, Any]:
        """Get kwargs for _setup_class call."""
        setup_class_kwargs = {
            "config": cls.configuration,
        }
        return setup_class_kwargs

    @classmethod
    def _build_image(cls) -> DockerImage:
        """Build the image."""
        client = docker.from_env()
        return ACNNodeDockerImage(client, config=cls.configuration)

    @classmethod
    def url(cls) -> str:
        """Get the url under which the image is reachable."""
        return str(cls.configuration.get("AEA_P2P_URI_PUBLIC"))


###
# Vanilla HardHat
###


@pytest.fixture(scope="session")
def hardhat_addr() -> str:
    """Get the hardhat addr"""
    return DEFAULT_HARDHAT_ADDR


@pytest.fixture(scope="session")
def hardhat_port() -> int:
    """Get the hardhat port"""
    return DEFAULT_HARDHAT_PORT


@pytest.fixture(scope="session")
def key_pairs() -> List[Tuple[str, str]]:
    """Get the default key paris for hardhat."""
    return KEY_PAIRS


class HardHatBaseTest(DockerBaseTest):
    """Base pytest class for HardHat."""

    addr: str = DEFAULT_HARDHAT_ADDR
    port: int = DEFAULT_HARDHAT_PORT

    @classmethod
    def setup_class_kwargs(cls) -> Dict[str, Any]:
        """Get kwargs for _setup_class call."""
        setup_class_kwargs = {
            "key_pairs": cls.key_pairs(),
            "url": cls.url(),
        }
        return setup_class_kwargs

    @classmethod
    def key_pairs(cls) -> List[Tuple[str, str]]:
        """Get the key pairs which are funded."""
        return KEY_PAIRS

    @classmethod
    def url(cls) -> str:
        """Get the url under which the image is reachable."""
        return f"{cls.addr}:{cls.port}"


###
# Custom HardHat: HardHat + Autonolas Registry
###


@pytest.fixture(scope="class")
def registries_scope_class(
    timeout: float = 2.0,
    max_attempts: int = 20,
) -> Generator:
    """Launch the Registry contracts image. This fixture is scoped to a class which means it will destroyed after running every test in a class."""
    client = docker.from_env()
    image = RegistriesDockerImage(client)
    yield from launch_image(image, timeout=timeout, max_attempts=max_attempts)


@pytest.mark.integration
class UseRegistries:
    """Inherit from this class to use a local Ethereum network with deployed registry contracts"""

    key_pairs: List[Tuple[str, str]] = KEY_PAIRS

    @classmethod
    @pytest.fixture(autouse=True)
    def _start_gnosis_and_registries(
        cls,
        registries_scope_class: Any,  # pylint: disable=redefined-outer-name
        hardhat_port: int,  # pylint: disable=redefined-outer-name
        key_pairs: List[Tuple[str, str]],  # pylint: disable=redefined-outer-name
    ) -> None:
        """Start a Hardhat instance, with registries contracts deployed."""
        cls.key_pairs = key_pairs


###
# Custom HardHat: HardHat + Gnosis Safe contracts
###


@pytest.fixture(scope="function")
def gnosis_safe_hardhat_scope_function(
    hardhat_addr: str,  # pylint: disable=redefined-outer-name
    hardhat_port: int,  # pylint: disable=redefined-outer-name
    timeout: float = 3.0,
    max_attempts: int = 40,
) -> Generator:
    """Launch the HardHat node with Gnosis Safe contracts deployed. This fixture is scoped to a function which means it will destroyed at the end of the test."""
    client = docker.from_env()
    logging.info(f"Launching Hardhat at port {hardhat_port}")
    image = GnosisSafeNetDockerImage(client, hardhat_addr, hardhat_port)
    yield from launch_image(image, timeout=timeout, max_attempts=max_attempts)


@pytest.fixture(scope="class")
def gnosis_safe_hardhat_scope_class(
    hardhat_addr: str,  # pylint: disable=redefined-outer-name
    hardhat_port: int,  # pylint: disable=redefined-outer-name
    timeout: float = 3.0,
    max_attempts: int = 40,
) -> Generator:
    """Launch the HardHat node with Gnosis Safe contracts deployed.This fixture is scoped to a class which means it will destroyed after running every test in a class."""
    client = docker.from_env()
    logging.info(f"Launching Hardhat at port {hardhat_port}")
    image = GnosisSafeNetDockerImage(client, hardhat_addr, hardhat_port)
    yield from launch_image(image, timeout=timeout, max_attempts=max_attempts)


@pytest.mark.integration
class UseGnosisSafeHardHatNet:
    """Inherit from this class to use HardHat local net with Gnosis-Safe deployed."""

    key_pairs: List[Tuple[str, str]] = KEY_PAIRS

    @classmethod
    @pytest.fixture(autouse=True)
    def _start_hardhat(
        cls,
        gnosis_safe_hardhat_scope_function: Any,  # pylint: disable=redefined-outer-name
        hardhat_port: int,  # pylint: disable=redefined-outer-name
        key_pairs: List[Tuple[str, str]],  # pylint: disable=redefined-outer-name
    ) -> None:
        """Start an HardHat instance."""
        cls.key_pairs = key_pairs


class HardHatGnosisBaseTest(HardHatBaseTest):
    """Base pytest class for HardHat with Gnosis deployed."""

    @classmethod
    def _build_image(cls) -> DockerImage:
        """Build the image."""
        client = docker.from_env()
        return GnosisSafeNetDockerImage(client, cls.addr, cls.port)


###
# Custom HardHat: HardHat + AMM contracts
###


class HardHatAMMBaseTest(HardHatBaseTest):
    """Base pytest class for HardHat with Gnosis and Uniswap deployed."""

    @classmethod
    def _build_image(cls) -> DockerImage:
        """Build the image."""
        client = docker.from_env()
        return AMMNetDockerImage(client, cls.addr, cls.port)


class RegistriesBaseTest(HardHatBaseTest):
    """Base pytest class for component registries."""

    addr: str = DEFAULT_HARDHAT_ADDR
    port: int = DEFAULT_HARDHAT_PORT

    @classmethod
    def _build_image(cls) -> DockerImage:
        """Build the image."""
        client = docker.from_env()
        return RegistriesDockerImage(
            client=client,
            addr=cls.addr,
            port=cls.port,
        )


@pytest.fixture(scope="class")
def ipfs_daemon() -> Iterator[bool]:
    """Starts an IPFS daemon for the tests."""
    print("Starting IPFS daemon...")
    daemon = IPFSDaemon()
    daemon.start()
    yield daemon.is_started()
    print("Tearing down IPFS daemon...")
    daemon.stop()


use_ipfs_daemon = pytest.mark.usefixtures("ipfs_daemon")
LOCAL_IPFS = "/dns/localhost/tcp/5001/http"


@pytest.fixture(scope="session")
def ipfs_domain() -> str:
    """Get the ipfs domain"""
    return LOCAL_IPFS


class UseLocalIpfs:
    """Use local IPFS daemon."""

    ipfs_domain: str

    @classmethod
    @pytest.fixture(autouse=True)
    def _start_ipfs_daemon(
        cls,
        ipfs_daemon: Any,  # pylint: disable=redefined-outer-name
        ipfs_domain: str,  # pylint: disable=redefined-outer-name
    ) -> None:
        """Start IPFS daemon."""
        cls.ipfs_domain = ipfs_domain



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Setup script for the plug-in."""

from setuptools import find_packages  # type: ignore
from setuptools import setup  # type: ignore


base_deps = [
    "open-aea[all]>=1.37.0,<2.0.0",
    "pytest==7.2.1",
    "open-aea-ledger-ethereum>=1.37.0,<2.0.0",
    "docker==6.1.2",
]

setup(
    name="open-aea-test-autonomy",
    version="0.10.10",
    author="Valory AG",
    license="Apache-2.0",
    description="Plugin containing test tools for open-autonomy packages.",
    long_description="Plugin containing test tools for open-autonomy packages.",
    long_description_content_type="text/markdown",
    packages=find_packages(
        where=".", include=["aea_test_autonomy", "aea_test_autonomy.*"]
    ),
    package_data={
        "aea_test_autonomy": [
            "py.typed",
            "data/*",
        ]
    },
    entry_points={},
    install_requires=base_deps,
    classifiers=[
        "Environment :: Console",
        "Environment :: Web Environment",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Communications",
        "Topic :: Internet",
        "Topic :: Software Development",
    ],
)
